<?php
# TagSoup_Parser - Lenient parser for HTML
# Version 0.2  Thu, 12 Apr 2007 20:34:24 +0000
#
# Copyright (C) 2007  Dwayne C. Litzenberger <dlitz@dlitz.net>
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Requires PHP >= 4.3.3
#
# Usage: Inherit from this base class, override handle_*(), then do something
# like this:
#  $rawdata = file_get_contents("input_file.html");
#  $p = new MyParser();
#  $p->parse($rawdata);
#
# Changelog:
# 0.1 - Sat, 20 Jan 2007 17:33:26 -0600 - Dwayne Litzenberger
#  * Initial Release
# 0.2 - Thu, 12 Apr 2007 20:34:24 +0000 - Dwayne Litzenberger
#  * Fix handling of decimal character entities (e.g. "&#37;")

class TagSoup_Parser
{
    var $entities_table;
    var $entity_regex;

    function TagSoup_Parser()
    {
        // Populate the entities_table table
        $this->entities_table = $this->get_entities_table();

        // Set the entity reference regex.  $1 refers to the entity name.
        $this->entity_regex = '/&(#[0-9]+|#x[0-9a-f]+|[a-z][a-z0-9\-.]*(?<=[a-z0-9]));?/is';
    }

    function parse($rawdata)
    {
        // This function assumes ASCII/UTF-8 input, and outputs UTF-8.
        // For reference, see http://www.w3.org/MarkUp/SGML/sgml-lex/sgml-lex
        // TODO: This currently requires that the entire document be parsed at once,
        // but there's no fundamental reason why that has to be the case.
        $offset = 0;
        $state = 0;
        $tag_name = null;
        $is_endtag = null;
        $attributes = null;

        $this->handle_start();

        for (;;) {
            if ($state === 0) { // Plain text
                if (!preg_match('#[<>&]#s',
                    $rawdata, $m, PREG_OFFSET_CAPTURE, $offset))
                {
                    // Reached end of input; We're done.
                    $this->handle_text(substr($rawdata, $offset, strlen($rawdata) - $offset));
                    break;
                }

                $token = $m[0][0];
                $token_offset = $m[0][1];
                if ($token_offset > $offset) {
                    $this->handle_text(substr($rawdata, $offset, $token_offset - $offset));
                    $offset = $token_offset;
                }
                if ($token === '<') {   // opening a tag
                    if (substr($rawdata, $offset, 4) === '<!--') {
                        // HTML comment
                        if (!preg_match('#<!--(.*?)-->#s',
                            $rawdata, $m, PREG_OFFSET_CAPTURE, $offset))
                        {
                            // The rest of the document is a comment
                            $this->handle_comment(substr($rawdata, $offset+4));
                            $offset = strlen($rawdata);
                        } else {
                            $this->handle_comment($m[1][0]);
                            $offset += strlen($m[0][0]);
                        }
                    } elseif (substr($rawdata, $offset, 2) === '<!') {
                        // SGML declaration
                        if (!preg_match('#<!(.*?)>#s',
                            $rawdata, $m, PREG_OFFSET_CAPTURE, $offset))
                        {
                            // Closing '>' not found.  Treat the '<' as text
                            $this->handle_text('<');
                            $offset++;
                        } else {
                            $this->handle_sgmldec($m[1][0]);
                            $offset += strlen($m[0][0]);
                        }

                    } elseif (!preg_match('#<(/)?([a-z!][a-z0-9\-_.:]*)#is',
                        $rawdata, $m, PREG_OFFSET_CAPTURE, $offset)
                        or $m[0][1] !== $offset)
                    {
                        // An unescaped textual '<'
                        $this->handle_text('<');
                        $offset++;
                    } else {
                        $tag_name = $m[2][0];
                        $is_endtag = ($m[1][1] >= 0);
                        $attributes = array();
                        $offset += strlen($m[0][0]);
                        $state = 1;
                    }
                } elseif ($token === '>') {     // An unescaped textual '>'
                    $this->handle_text('>');
                    $offset++;
                } elseif ($token === '&') {     // An entity
                    if (!preg_match($this->entity_regex,
                        $rawdata, $m, PREG_OFFSET_CAPTURE, $offset)
                        or $m[0][1] !== $offset)
                    {
                        // An unescaped textual '&'
                        $this->handle_text('&');
                        $offset++;
                    } else {
                        $replacement = $this->decode_entity($m[1][0]);
                        if ($replacement === false) {
                            // The entity doesn't decode, so treat the & as an
                            // unescaped textual '&'
                            $this->handle_text('&');
                            $offset++;
                        } else {
                            $this->handle_text($replacement);
                            $offset += strlen($m[0][0]);
                        }
                    }
                } else {    // this should never happen
                    die("BUG in line ".__LINE__." of ".basename(__FILE__));
                }

            } elseif ($state === 1) {   // Inside a tag
                if (!preg_match(
                    '#(>)|(<)|\s+([a-z_][a-z0-9\-_:]*)(\s*=\s*(?:'.
                    '"([^"]*)"|\'([^\']*)\'|([^\s>]*)'.
                    '))?#is',
                    $rawdata, $m, PREG_OFFSET_CAPTURE, $offset))
                {
                    if ($is_endtag) {
                        $this->handle_endtag($tag_name);
                    } else {
                        $this->handle_starttag($tag_name, $attributes);
                    }
                    $state = 0;
                } else {
                    if ($m[1][1] >= 0 or $m[2][1] >= 0) {     // Reached '>' or '<'
                        if ($is_endtag) {
                            $this->handle_endtag($tag_name);
                        } else {
                            $this->handle_starttag($tag_name, $attributes);
                        }
                        $offset = $m[0][1];
                        if ($m[1][1] >= 0) {    // Reached '>'
                            $offset++;
                        }
                        $state = 0;
                    } else {    // attribute
                        $attr_name = $m[3][0];
                        if ($m[4][1] < 0) {    // no attribute value
                            $attr_value = true;
                        } elseif ($m[5][1] >= 0) {  // double-quoted attr value
                            $attr_value = $this->decode_entities($m[5][0]);
                        } elseif ($m[6][1] >= 0) {  // single-quoted attr value
                            $attr_value = $this->decode_entities($m[6][0]);
                        } elseif ($m[7][1] >= 0) {  // unquoted attribute value
                            $attr_value = $this->decode_entities($m[7][0]);
                        } else {
                            die("BUG in line ".__LINE__." of ".basename(__FILE__));
                        }
                        if (!isset($attributes[$attr_name])) {
                            $attributes[$attr_name] = $attr_value;
                        }
                        $offset = $m[0][1] + strlen($m[0][0]);
                    }
                }
            } else {    // Invalid value for $state
                die("BUG in line ".__LINE__." of ".basename(__FILE__));
            }
        }
        
        $this->handle_finish();
    }

    function decode_entity($name)
    {
        if (substr($name, 0, 1) === '#') { // Numeric character entity reference
            $tmp = substr($name, 1, 1);
            if ($tmp === 'x' or $tmp === 'X') {
                // Hexadecimal character entity reference
                return $this->utf8chr(hexdec(substr($name, 2)));
            } else {
                // Decimal character entity reference
                return $this->utf8chr(intval(substr($name, 1)));
            }
        }
        if (isset($this->entities_table[$name])) {
            return $this->entities_table[$name];
        } else {
            return false;
        }
    }

    function decode_entities($rawdata)
    {
        $retval = "";
        $offset = 0;
        while (preg_match($this->entity_regex,
                $rawdata, $m, PREG_OFFSET_CAPTURE, $offset))
        {
            $match_length = strlen($m[0][0]);
            $match_offset = $m[0][1];
            $entity_name = $m[1][0];
            if ($match_offset !== $offset) {
                $retval .= substr($rawdata, $offset, $match_offset - $offset);
                $offset = $match_offset;
            }
            
            $replacement = $this->decode_entity($entity_name);
            if ($replacement === false) {
                // The entity doesn't decode, so treat the & as an unescaped
                // textual '&'
                $retval .= '&';
                $offset++;
            } else {
                $retval .= $replacement;
                $offset += $match_length;
            }
        }
        $retval .= substr($rawdata, $offset);
        return $retval;
    }

    function utf8chr($num)
    {
        // Return UTF-8 character by unicode code point.  See RFC 3629.
        // Done manually so that we don't have to rely on mbstring being
        // set up properly.
        if (!isset($num) or !is_numeric($num) or $num < 0 or $num > 0x10FFFF) {
            // Code point is out of range
            trigger_error("utf8chr() argument not numeric in range [0, 0x10FFFF]", E_USER_NOTICE);
            return '\xef\xbf\xbd';  // U+FFFD
        }
        $num = intval($num);
        if ($num <= 0x7F) {
            return chr($num);
        } elseif ($num <= 0x07FF) {
            return chr(0xc0 | ($num >> 6)) .
                   chr(0x80 | ($num & 0x3f));
        } elseif ($num <= 0xFFFF) {
            return chr(0xe0 | ($num >> 12)) .
                   chr(0x80 | (($num >> 6) & 0x3f)) .
                   chr(0x80 | ($num & 0x3f));
        } else {    // $num <= 0x10FFFF
            return chr(0xf0 | ($num >> 18)) .
                   chr(0x80 | (($num >> 12) & 0x3f)) .
                   chr(0x80 | (($num >> 6) & 0x3f)) .
                   chr(0x80 | ($num & 0x3f));
        }
    }

    function get_entities_table()
    {
        // Generate the UTF-8 entities table from the HTML4 code-point table
        $table = array();
        foreach ($this->get_html4_entities_table() as $name => $codepoint) {
            $table[$name] = $this->utf8chr($codepoint);
        }
        
        // Add the XML "apos" character entity, which is not included in HTML 4
        $table['apos'] = "'";
        
        // Add the all-caps character entities that browsers seem to tolerate
        $caps_list = array('quot', 'amp', 'lt', 'gt', 'copy', 'reg', 'trade');
        foreach ($caps_list as $name) {
            $table[strtoupper($name)] = $table[$name];
        }
        
        return $table;
    }

    function get_html4_entities_table()
    {
        // This associative array maps the 252 HTML 4 character entity
        // references to their Unicode code points.
        // See http://www.w3.org/TR/html4/sgml/entities.html
        // Note: As of this writing, PHP's get_html_translation_table()
        // function can't be used to generate this entire table. (It only
        // supports characters in the ISO-8859-1 character set.)
        return array(
            'quot' => 0x22, 'amp' => 0x26, 'lt' => 0x3C, 'gt' => 0x3E, 
            'nbsp' => 0xA0, 'iexcl' => 0xA1, 'cent' => 0xA2, 'pound' => 0xA3, 
            'curren' => 0xA4, 'yen' => 0xA5, 'brvbar' => 0xA6, 'sect' => 0xA7, 
            'uml' => 0xA8, 'copy' => 0xA9, 'ordf' => 0xAA, 'laquo' => 0xAB, 
            'not' => 0xAC, 'shy' => 0xAD, 'reg' => 0xAE, 'macr' => 0xAF, 
            'deg' => 0xB0, 'plusmn' => 0xB1, 'sup2' => 0xB2, 'sup3' => 0xB3, 
            'acute' => 0xB4, 'micro' => 0xB5, 'para' => 0xB6, 'middot' => 0xB7, 
            'cedil' => 0xB8, 'sup1' => 0xB9, 'ordm' => 0xBA, 'raquo' => 0xBB, 
            'frac14' => 0xBC, 'frac12' => 0xBD, 'frac34' => 0xBE, 
            'iquest' => 0xBF, 'Agrave' => 0xC0, 'Aacute' => 0xC1, 
            'Acirc' => 0xC2, 'Atilde' => 0xC3, 'Auml' => 0xC4, 'Aring' => 0xC5, 
            'AElig' => 0xC6, 'Ccedil' => 0xC7, 'Egrave' => 0xC8, 
            'Eacute' => 0xC9, 'Ecirc' => 0xCA, 'Euml' => 0xCB, 
            'Igrave' => 0xCC, 'Iacute' => 0xCD, 'Icirc' => 0xCE, 
            'Iuml' => 0xCF, 'ETH' => 0xD0, 'Ntilde' => 0xD1, 'Ograve' => 0xD2, 
            'Oacute' => 0xD3, 'Ocirc' => 0xD4, 'Otilde' => 0xD5, 
            'Ouml' => 0xD6, 'times' => 0xD7, 'Oslash' => 0xD8, 
            'Ugrave' => 0xD9, 'Uacute' => 0xDA, 'Ucirc' => 0xDB, 
            'Uuml' => 0xDC, 'Yacute' => 0xDD, 'THORN' => 0xDE, 'szlig' => 0xDF, 
            'agrave' => 0xE0, 'aacute' => 0xE1, 'acirc' => 0xE2, 
            'atilde' => 0xE3, 'auml' => 0xE4, 'aring' => 0xE5, 'aelig' => 0xE6, 
            'ccedil' => 0xE7, 'egrave' => 0xE8, 'eacute' => 0xE9, 
            'ecirc' => 0xEA, 'euml' => 0xEB, 'igrave' => 0xEC, 
            'iacute' => 0xED, 'icirc' => 0xEE, 'iuml' => 0xEF, 'eth' => 0xF0, 
            'ntilde' => 0xF1, 'ograve' => 0xF2, 'oacute' => 0xF3, 
            'ocirc' => 0xF4, 'otilde' => 0xF5, 'ouml' => 0xF6, 
            'divide' => 0xF7, 'oslash' => 0xF8, 'ugrave' => 0xF9, 
            'uacute' => 0xFA, 'ucirc' => 0xFB, 'uuml' => 0xFC, 
            'yacute' => 0xFD, 'thorn' => 0xFE, 'yuml' => 0xFF, 
            'OElig' => 0x152, 'oelig' => 0x153, 'Scaron' => 0x160, 
            'scaron' => 0x161, 'Yuml' => 0x178, 'fnof' => 0x192, 
            'circ' => 0x2C6, 'tilde' => 0x2DC, 'Alpha' => 0x391, 
            'Beta' => 0x392, 'Gamma' => 0x393, 'Delta' => 0x394, 
            'Epsilon' => 0x395, 'Zeta' => 0x396, 'Eta' => 0x397, 
            'Theta' => 0x398, 'Iota' => 0x399, 'Kappa' => 0x39A, 
            'Lambda' => 0x39B, 'Mu' => 0x39C, 'Nu' => 0x39D, 'Xi' => 0x39E, 
            'Omicron' => 0x39F, 'Pi' => 0x3A0, 'Rho' => 0x3A1, 
            'Sigma' => 0x3A3, 'Tau' => 0x3A4, 'Upsilon' => 0x3A5, 
            'Phi' => 0x3A6, 'Chi' => 0x3A7, 'Psi' => 0x3A8, 'Omega' => 0x3A9, 
            'alpha' => 0x3B1, 'beta' => 0x3B2, 'gamma' => 0x3B3, 
            'delta' => 0x3B4, 'epsilon' => 0x3B5, 'zeta' => 0x3B6, 
            'eta' => 0x3B7, 'theta' => 0x3B8, 'iota' => 0x3B9, 
            'kappa' => 0x3BA, 'lambda' => 0x3BB, 'mu' => 0x3BC, 'nu' => 0x3BD, 
            'xi' => 0x3BE, 'omicron' => 0x3BF, 'pi' => 0x3C0, 'rho' => 0x3C1, 
            'sigmaf' => 0x3C2, 'sigma' => 0x3C3, 'tau' => 0x3C4, 
            'upsilon' => 0x3C5, 'phi' => 0x3C6, 'chi' => 0x3C7, 'psi' => 0x3C8, 
            'omega' => 0x3C9, 'thetasym' => 0x3D1, 'upsih' => 0x3D2, 
            'piv' => 0x3D6, 'ensp' => 0x2002, 'emsp' => 0x2003, 
            'thinsp' => 0x2009, 'zwnj' => 0x200C, 'zwj' => 0x200D, 
            'lrm' => 0x200E, 'rlm' => 0x200F, 'ndash' => 0x2013, 
            'mdash' => 0x2014, 'lsquo' => 0x2018, 'rsquo' => 0x2019, 
            'sbquo' => 0x201A, 'ldquo' => 0x201C, 'rdquo' => 0x201D, 
            'bdquo' => 0x201E, 'dagger' => 0x2020, 'Dagger' => 0x2021, 
            'bull' => 0x2022, 'hellip' => 0x2026, 'permil' => 0x2030, 
            'prime' => 0x2032, 'Prime' => 0x2033, 'lsaquo' => 0x2039, 
            'rsaquo' => 0x203A, 'oline' => 0x203E, 'frasl' => 0x2044, 
            'euro' => 0x20AC, 'image' => 0x2111, 'weierp' => 0x2118, 
            'real' => 0x211C, 'trade' => 0x2122, 'alefsym' => 0x2135, 
            'larr' => 0x2190, 'uarr' => 0x2191, 'rarr' => 0x2192, 
            'darr' => 0x2193, 'harr' => 0x2194, 'crarr' => 0x21B5, 
            'lArr' => 0x21D0, 'uArr' => 0x21D1, 'rArr' => 0x21D2, 
            'dArr' => 0x21D3, 'hArr' => 0x21D4, 'forall' => 0x2200, 
            'part' => 0x2202, 'exist' => 0x2203, 'empty' => 0x2205, 
            'nabla' => 0x2207, 'isin' => 0x2208, 'notin' => 0x2209, 
            'ni' => 0x220B, 'prod' => 0x220F, 'sum' => 0x2211, 
            'minus' => 0x2212, 'lowast' => 0x2217, 'radic' => 0x221A, 
            'prop' => 0x221D, 'infin' => 0x221E, 'ang' => 0x2220, 
            'and' => 0x2227, 'or' => 0x2228, 'cap' => 0x2229, 'cup' => 0x222A, 
            'int' => 0x222B, 'there4' => 0x2234, 'sim' => 0x223C, 
            'cong' => 0x2245, 'asymp' => 0x2248, 'ne' => 0x2260, 
            'equiv' => 0x2261, 'le' => 0x2264, 'ge' => 0x2265, 'sub' => 0x2282, 
            'sup' => 0x2283, 'nsub' => 0x2284, 'sube' => 0x2286, 
            'supe' => 0x2287, 'oplus' => 0x2295, 'otimes' => 0x2297, 
            'perp' => 0x22A5, 'sdot' => 0x22C5, 'lceil' => 0x2308, 
            'rceil' => 0x2309, 'lfloor' => 0x230A, 'rfloor' => 0x230B, 
            'lang' => 0x2329, 'rang' => 0x232A, 'loz' => 0x25CA, 
            'spades' => 0x2660, 'clubs' => 0x2663, 'hearts' => 0x2665, 
            'diams' => 0x2666,
        );
    }
    
    //
    // The following methods should be overridden by the user
    //

    // handle_start - called when parsing starts
    function handle_start() { }

    // handle_text - called when the parser encounters text.
    function handle_text($content) { }

    // handle_starttag - called when the parser encounters a starting tag, e.g.
    // "<p>"
    function handle_starttag($tagname, $attributes=null) { } 

    // handle_starttag - called when the parser encounters an ending tag, e.g.
    // "</p>"
    function handle_endtag($tagname) { }
    
    // handle_comment - called when the parser encounters a comment
    function handle_comment($content) { }

    // handle_sgmldec - called when the parser encounters an SGML declaration,
    // e.g. <!DOCTYPE...>
    function handle_sgmldec($content) { }

    // handle_finish - called when parsing is finished
    function handle_finish() { }
}
