<?php
# TagSoup_ToXml - Generate well-formed XML from HTML tag soup
# Version 0.1  Sat, 20 Jan 2007 17:33:26 -0600
#
# Copyright (C) 2007  Dwayne C. Litzenberger <dlitz@dlitz.net>
# 
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Requires PHP >= 4.3.3

require_once "TagSoup/Parser.php";

class TagSoup_ToXml extends TagSoup_Parser
{
    // Public
    var $omit_comments;
    
    // Private
    var $forbidden_endtags;
    var $forbidden_parents;
    var $mandatory_parents;
    
    var $tagstack;
    
    var $prologue;
    var $html_tag;
    var $head_tag;
    var $head_content;
    var $body_tag;
    var $body_content;

    function TagSoup_ToXml()
    {
        $this->TagSoup_Parser();

        $this->tagstack = array();
        $this->omit_comments = false;
        $this->omit_sgmldecs = false;

        // Elements that are always empty, according to the HTML4 spec:
        //  http://www.w3.org/TR/html4/index/elements.html
        $this->forbidden_endtags = array(
            'area', 'base', 'basefont', 'br', 'col', 'frame', 'hr', 'img',
            'input', 'isindex', 'link', 'meta', 'param');
        
        // Elements that should not be their own parents.
        // There are probably bugs here.
        $this->forbidden_parents = array('p', 'option', 'col', 'colgroup');
        
        // Elements that should have specific parents (note that this does not
        // actually guarantee that these elements will have one of the
        // specified parents).
        // There are probably bugs here.
        $this->mandatory_parents = array(
            'li' => array('ul', 'ol'),
            'td' => array('tr', 'table'),
            'th' => array('tr', 'table'),
            'tr' => array('table', 'tbody', 'thead', 'tfoot'),
            'dd' => array('dl'),
            'dt' => array('dl'),
        );

        $this->output_buffer = null;

        $this->prologue = "";
        $this->html_tag = null;
        $this->head_tag = null;
        $this->head_content = null;
        $this->body_tag = null;
        $this->body_content = null;
    }

    function add_output($text) {
        if (!is_null($this->output_buffer)) {
            $this->output_buffer .= $text;
        } elseif (is_null($this->body_content)) {
            $this->head_content .= $text;
        } else {
            $this->body_content .= $text;
        }
    }

    function handle_text($text)
    {
        $this->add_output(htmlspecialchars($text));
    }

    function handle_starttag($tagname, $attributes=null)
    {
        // Lowercase the tag name
        $tagname = strtolower($tagname);

        // Lowercase and sort the attribute names
        if (is_array($attributes)) {
            $attrs = array();
            foreach ($attributes as $k => $v) {
                $attrs[strtolower($k)] = $v;
            }
            $attributes = $attrs;
            unset($attrs);
            ksort($attributes);
        }

        // Ignore duplicate HTML, HEAD, and BODY tags
        if (($tagname === "html" and !is_null($this->html_tag)) or
            ($tagname === "head" and !is_null($this->head_tag)) or
            ($tagname === "body" and !is_null($this->body_tag)))
        {
            return;
        }
        
        // Handle elements that shouldn't be parents of themselves
        foreach ($this->forbidden_parents as $tn) {
            if ($tagname === $tn
                and $this->tagstack[count($this->tagstack)-1] === $tn)
            {
                $this->add_output("</$tn>");
                array_pop($this->tagstack);
                break;
            }
        }
        
        // Handle tags that should have specific parents
        foreach ($this->mandatory_parents as $child => $parents) {
            if ($tagname === $child) {    // li should have ul or ol as a parent
                for ($i = count($this->tagstack)-1; $i >= 0; $i--) {
                    if (in_array($this->tagstack[$i], $parents)) {
                        for ($j = count($this->tagstack)-1; $j > $i; $j--) {
                            $tn = array_pop($this->tagstack);
                            $this->add_output("</$tn>");
                        }
                        break;
                    }
                }
                break;
            }
        }

        if ($tagname === "html" or $tagname === "head") {
            $this->output_buffer = "";
        } elseif ($tagname === "body") {
            $this->handle_endtag("head");
            $this->output_buffer = "";
        }

        $this->add_output("<${tagname}");
        if (is_array($attributes)) {
            foreach ($attributes as $k => $v) {
                $this->add_output(' '.$k.'="'.htmlspecialchars($v).'"');
            }
        }
        if (in_array($tagname, $this->forbidden_endtags)) {
            $this->add_output(" />"); // Self-closing tag
        } else {
            $this->add_output(">");
            if ($tagname !== "html" and $tagname !== "head" and $tagname !== "body") {
                array_push($this->tagstack, $tagname);
            }
        }

        if ($tagname === "html") {
            $this->html_tag = $this->output_buffer;
            $this->output_buffer = null;
        } elseif ($tagname === "head") {
            $this->head_tag = $this->output_buffer;
            $this->output_buffer = null;
        } elseif ($tagname === "body") {
            $this->body_tag = $this->output_buffer;
            $this->output_buffer = null;
            if (is_null($this->body_content)) {
                $this->body_content = "";
            }
        }
    } 

    function handle_endtag($tagname)
    {
        // Lowercase the tag name
        $tagname = strtolower($tagname);
        
        // Don't close HTML, HEAD, or BODY tags here, but close everything else.
        if ($tagname === "html" or $tagname === "head" or $tagname === "body") {
            while (count($this->tagstack) > 0) {
                $tn = array_pop($this->tagstack);
                $this->add_output("</${tn}>");
            }
            return;
        }

        if (!in_array($tagname, $this->tagstack)) {
            // If the tag wasn't opened, don't close it.
            return;
        }
        
        // Force proper nesting of tags
        while ($this->tagstack[count($this->tagstack)-1] !== $tagname
            and count($this->tagstack) > 0)
        {
            $tn = array_pop($this->tagstack);
            $this->add_output("</${tn}>");
        }
        if (count($this->tagstack) == 0) {  // This should never happen
            die("BUG in line ".__LINE__." of ".basename(__FILE__));
        }

        // Close the tag
        $tn = array_pop($this->tagstack);
        if ($tn !== $tagname) {
            die("BUG in line ".__LINE__." of ".basename(__FILE__));
        }
        $this->add_output("</${tagname}>");
    }
    
    function handle_comment($string)
    {
        if ($this->omit_comments) {
            return;
        }
        
        // According to XML 1.0 section 2.5 ("Comments"):
        //  "For compatibility, the string "--" (double-hyphen) MUST NOT occur
        //  within comments."
        $string = preg_replace('#-(?=-)#s', '- ', $string);

        // Also, XML comments can't end in '-'
        if (substr($string, strlen($string)-1) === '-') {
            $string .= ' ';
        }

        $this->add_output("<!--${string}-->");
    }

    function handle_sgmldec($string)
    {
        return;     // Omit SGML declarations
    }

    function handle_start()
    {
        // Do nothing
    }
    
    function handle_finish()
    {
        // Close all open tags
        while (count($this->tagstack) > 0)
        {
            $tn = array_pop($this->tagstack);
            $this->add_output("</${tn}>");
        }

        if (is_null($this->html_tag)) {
            $this->html_tag = "<html>";
        }
        if (is_null($this->head_tag)) {
            $this->head_tag = "<head>";
        }
        if (is_null($this->body_tag)) {
            $this->body_tag = "<body>";
            $this->body_content = $this->head_content;
            $this->head_content = "";
        }

        // Dump output
        print $this->prologue;
        print $this->html_tag;
        print $this->head_tag;
        print $this->head_content;
        print "</head>";
        print $this->body_tag;
        print $this->body_content;
        print "</body></html>";
    }
}
